% SIMULATION OF THE HAUT GLACIER D'AROLLA VELOCITY FIELD
% No Slip Boundary Condition at the base and Fixed Ice Surface
%
% The underlying model is the First Order Stokes Model (Blatter-Pattyn)
%
% The numerical method is the Global RBF Method

% If you use this code please cite JOSEFIN AHLKRONA and VICTOR SHCHERBAKOV
% "A MESHFREE APPROACH TO NON-NEWTONIAN FREE SURFACE ICE FLOW:
% APPLICATION TO THE HAUT GLACIER D'AROLLA"


% --- (C) Josefin Ahlkrona & Victor Shcherbakov 2016 ---

function [vel_x,vel_y,Niter] = DArollaNoSlipRBF(Nx,ep,phi,tol,kmax)


%% Data

% Load the Haut Glacier d'Arolla profile
% top is surface coordinates, bot is base coordinates
% pol = [bot; top] is full profile

load('DArollaProfile');

% Scale down from meters to kilometers
pol = pol/1000;
top = top/1000;
bot = bot/1000;

% Define density, gravity, and rate factor
rho = 9.03721418e-13;
g = 9.7692e+12;
Afact = 100;


%% Background grid

% Define boundaries of the grid
xmax = max(pol(:,1));
xmin = min(pol(:,1));
zmax = max(pol(:,2));
zmin = min(pol(:,2));

% Select number of points used in x-direction    
hx = (xmax-xmin)/Nx;

% Nx/4 number of points is used in z-direction
Nz = floor(Nx/4);
hz = (zmax-zmin)/Nz;

% Fill distnace 
hfill = min(hx,hz);

% Define the grid
[X,Y] = meshgrid(xmin:hx:xmax,zmin:hz:zmax);
xgrid = [X(:) Y(:)];


%% Refine the boundary

% Since the original data for the Haut Glacier d'Arolla
% has step 100m and contains only 50 points for the surface and 
% 50 points for the base, we would like to refine it to have the same
% number of points and in x-direction

% Refined x-direction
xc = xmin:hx:xmax;

% Old boundary coordinates contain 50 points
xbot = bot;
xsurf = top;

% New boundary coordinates contain Nx points
xbotnew = interp1(xbot(:,1),xbot(:,2),xc,'spline');
xsurfnew = interp1(xsurf(:,1),xsurf(:,2),xc,'spline');

xbotnew = [xc(2:end-1)' xbotnew(2:end-1)'];
xsurfnew = [xc' xsurfnew'];

% xb defines the full refined boundary profile
xb = [xbotnew; xsurfnew];


%% Determine internal points

% Find which points fall inside the domain
% Remove the points which lie outside
% xi denotes internal points

pol_new = [[0 3.2]; xbotnew; flipud(xsurfnew)];
inPol = inpolygon(xgrid(:,1),xgrid(:,2),pol_new(:,1),pol_new(:,2));
inPol = find(inPol);
xi = xgrid(inPol,:);


%% Determine indistinct points

% Identify points which fall in indistinct_par-neighbourhood of each other
% and leave just one copy. Otherwise RBF matrices will be ill-conditioned

% This procedure is necessary because, when adding the boundary points to
% the grid, some grid points may coincide with the boundary. This leads to
% singularities in RBF matrices

indistinct_par = hfill/2;

indremove = [];
indkeep = [];
for i = 1:length(xi)
    for j = 1:length(xb)
        indistinct(j) = norm(xi(i,:)-xb(j,:));
    end
    indistinct_sort = sort(indistinct);
    if indistinct_sort(1) < indistinct_par
        indremove = [indremove i];
    else
        indkeep = [indkeep i];
    end
end

% Plot the boundary profile, the background grid, the internal points, and
% the removed indistinct points
figure;
hold on
plot(xgrid(:,1),xgrid(:,2),'bs');
plot(pol_new(:,1),pol_new(:,2),'ko-','linewidth',3);
plot(xi(:,1),xi(:,2),'rs');
plot(xi(indremove,1),xi(indremove,2),'ms','markersize',8,'markerfacecolor','m')
axis tight
legend('Background nodes','Boundary nodes', 'Internal nodes', 'Indistinct nodes')
set(gca,'fontsize',20)
drawnow

% Remove the indistinct nodes
xi = xi(indkeep,:);


%% Construct the set of computational points

% We combine the internal and the boundary nodes into one set of nodes
% This set does not represent a grid any more, because the boundary nodes
% do not belong to the grid

% The set of points
xe = [xi;xb];

% The number of computational points
Ne = length(xe);

% Define indices which correspond to internal and boundary point in the set   
ind_inter = 1:length(xi); % internal points
ind_bound = length(xi)+1:length(xe); % all boundary points
ind_bot = length(xi)+1:length(xi)+length(xbotnew); % bottom points
ind_surf = length(xi)+length(xbotnew)+1:length(xe); % surface points


%% Compute the derivative of the surface

% The surface derivative is needed for evaluation the horizontal velocity
% field of ice. We use RBF approximation to calculate the derivative

% Find distances between the points on the surface
rc = xcdist(xc',xc','0');

% Compute RBF matrices
Ac = RBFmat(phi,ep,rc,'0',1);
Axc = RBFmat(phi,ep,rc,'1',1)/Ac;

% Compute the derivative
surfder = Axc*xsurfnew(:,2);
surfder = surfder';

% Assign the value of the surface derivative to the corresponding internal
% points
surfderField = repmat(surfder,Nz+1,1);
surfderField = surfderField(:);
surfderField = surfderField(inPol);
surfderField = surfderField(indkeep);

% Define a vector which depends on the surface derivative and comes in the
% right hand side
surfderField_rhs = [surfderField; zeros(length(xb),1)];

surfder = fliplr(surfder);


%% RBF matrices

% Find distances between points in the set
rc = xcdist(xe,xe,'0');

% Compute RBF martices
A0 = RBFmat(phi,ep,rc,'0',1); % interpolation matrix
Ax = RBFmat(phi,ep,rc,'1',1)/A0; % first x-derivative
Ay = RBFmat(phi,ep,rc,'1',2)/A0; % first y-derivative
Axx = RBFmat(phi,ep,rc,'2',1)/A0; % second x-derivative
Axy = RBFmat(phi,ep,rc,'m2',[1,2])/A0; % mixed derivative
Ayy = RBFmat(phi,ep,rc,'2',2)/A0; % second y-derivative


%% Nonlenear solver -- Fixed point iteration 

% In the nonlinear itration we solve for the horizontal velocity

% Make an initial guess
uold = 0.07*ones(length(xe),1);

% Define the counter
k = 1;

err = 1;
while err > tol
    
    % Compute derivatives of the horizontal velocity as 
    % it comes as a multiplier in the viscosity
    dudx = Ax*uold;
    dudy = Ay*uold; 
    
    % Set a critical shear rate
    crit_shear_rate = 10^(-10);
    
    % Compute the viscosity eta
    eta = 0.5*Afact^(-1/3)*(dudx.^2 + 0.25*dudy.^2).^(-1/3);
    
    % Find where the viscosity is smaller than the critical shear rate
    ind_eta_zero = find(abs(dudx.^2 + 0.25*dudy.^2)<crit_shear_rate);
    
    % Replace by the critical shear rate, otherwise singularity
    eta(ind_eta_zero) = 0.5*Afact^(-1/3)*(crit_shear_rate).^(-1/3);
    
    % Compute the gradient of the viscosity
    grad_eta_x = -Afact^(-1/3)*((dudx.^2 + 0.25*dudy.^2).^(-4/3))/6 .* ...
        (2*dudx.*(Axx*uold) + 0.5*dudy.*(Axy*uold)); 
    grad_eta_x(ind_eta_zero) = 0; 
    
    grad_eta_y = -Afact^(-1/3)*((dudx.^2 + 0.25*dudy.^2).^(-4/3))/6 .* ...
        (2*dudx.*(Axy*uold) + 0.5*dudy.*(Ayy*uold)); 
    grad_eta_y(ind_eta_zero) = 0; 

    % Make an exception for the first iteration
    if k == 1
        eta = ones(size(eta));
        
        grad_eta_x = zeros(size(eta));
        grad_eta_y = zeros(size(eta));
    end
    
    % Define the discrete spatial differential operator
    L = 4*diag(grad_eta_x)*Ax + 1*diag(grad_eta_y)*Ay + diag(eta)*(4*Axx + 1*Ayy);

    % Define the discrete boundary differential operator at the surface
    Lsurf = -2*diag(surfder)*Ax(ind_surf,:) + 0.5*Ay(ind_surf,:);

    % Define the discrete boundary operator at the base
    % Accorsing to the no slip boundary it is an identity operator
    Lbot = eye(length(ind_bot));

    % Embed these boundary parts into the global system
    L(ind_surf,:) = Lsurf;
    L(ind_bot,:) = 0;
    L(ind_bot,ind_bot) = Lbot;

    % Define the right hand side
    rhs = rho*g*surfderField_rhs;

    % Perform an LU decomposition for a more efficient matrix inversion
    [Lx,Ux] = lu(L);

    % Solve for the horizontal velocity
    vel_x = Ux\(Lx\rhs);
    
    % Compute the error between old and new nonlinear iterations
    err = rms(uold-vel_x);
    
    % Update
    uold = vel_x;

    % Update the counter
    k = k + 1;
    
    % If not converged but exceeded the maximal number of iteration, break 
    if k > kmax
        break
    end

end

% Number of taken iterations
Niter = k-1;

%% Solve for the vertical velocity

% Define the differntial operator to solve for the vertical velocity
By = Ay;
By(ind_bot,:) = 0;
Bbot = eye(length(ind_bot));
By(ind_bot,ind_bot) = Bbot;

% Compute x-derivative of the horizontal velocity
vel_x_der = Ax*vel_x;

% Incorporate the boundary condititon
vel_x_der(ind_bot) = 0;

% Perform an LU decomposition for a more efficient matrix inversion
[Ly,Uy] = lu(By);

% Solve for the vertical velocity
vel_y = -Uy\(Ly\vel_x_der);


%% Visualize the results

% flip the boundary points around so that it is one connected line
% otherwise the triangulation routine is unhappy. 
xtri=xe; utri=vel_x;
xtrib=xe(ind_bound,:); utrib=vel_x(ind_bound,:);
xtrb1=xtrib(1:length(ind_bound)/2-1,:); utrb1=utrib(1:length(ind_bound)/2-1,:);
xtrb2=xtrib(length(ind_bound)/2:end,:); utrb2=utrib(length(ind_bound)/2:end,:);
xtri(ind_bound,:)=[xtrb1; flipud(xtrb2)]; utri(ind_bound,:)=[utrb1; flipud(utrb2)];

% find triangulation that the solution will be interpolated to
Tri= findTriangulation(xtri(:,1),xtri(:,2),ind_bound);

% Finally, interpolate and the plot of the velocity. It looks smooth to me.
figure;
trisurf(Tri,xtri(:,1),xtri(:,2),utri); view(2);
shading interp;
colormap('cool')
axis tight
colorbar
title('Horizontal velocity')
set(gca,'fontsize',20)


% flip the boundary points around so that it is one connected line
% otherwise the triangulation routine is unhappy. 
wtri=vel_y;
wtrib=vel_y(ind_bound,:);
wtrb1=wtrib(1:length(ind_bound)/2-1,:);
wtrb2=wtrib(length(ind_bound)/2:end,:);
wtri(ind_bound,:)=[wtrb1; flipud(wtrb2)];

% Finally, interpolate and the plot of the velocity. It looks smooth to me.
figure;
trisurf(Tri,xtri(:,1),xtri(:,2),wtri); view(2);
shading interp;
colormap('cool')
axis tight
colorbar
title('Vertical velocity')
set(gca,'fontsize',20)






