% SIMULATION OF THE HAUT GLACIER D'AROLLA VELOCITY FIELD
% No Slip Boundary Condition at the base and Fixed Ice Surface
%
% The underlying model is the First Order Stokes Model (Blatter-Pattyn)
%
% The numerical method is RBF-PUM

% If you use this code please cite JOSEFIN AHLKRONA and VICTOR SHCHERBAKOV
% "A MESHFREE APPROACH TO NON-NEWTONIAN FREE SURFACE ICE FLOW:
% APPLICATION TO THE HAUT GLACIER D'AROLLA"


% --- (C) Josefin Ahlkrona & Victor Shcherbakov 2016 ---

function [vel_x,Niter] = DArollaNoSlipRBFPUM(Nx,ep,phi,tol,kmax,R)


%% Data

% Load the Haut Glacier d'Arolla profile
% top is surface coordinates, bot is base coordinates
% pol = [bot; top] is full profile

load('DArollaProfile');

% Scale down from meters to kilometers
pol = pol/1000;
top = top/1000;
bot = bot/1000;

% Define density, gravity, and rate factor
rho = 9.03721418e-13;
g = 9.7692e+12;
Afact = 100;


%% Background grid

% Define boundaries of the grid
xmax = max(pol(:,1));
xmin = min(pol(:,1));
zmax = max(pol(:,2));
zmin = min(pol(:,2));

% Select number of points used in x-direction    
hx = (xmax-xmin)/Nx;

% Nx/4 number of points is used in z-direction
Nz = floor(Nx/4);
hz = (zmax-zmin)/Nz;

% Fill distnace 
hfill = min(hx,hz);

% Define the grid
[X,Y] = meshgrid(xmin:hx:xmax,zmin:hz:zmax);
xgrid = [X(:) Y(:)];


%% Refine the boundary

% Since the original data for the Haut Glacier d'Arolla
% has step 100m and contains only 50 points for the surface and 
% 50 points for the base, we would like to refine it to have the same
% number of points and in x-direction

% Refined x-direction
xc = xmin:hx:xmax;

% Old boundary coordinates contain 50 points
xbot = bot;
xsurf = top;

% New boundary coordinates contain Nx points
xbotnew = interp1(xbot(:,1),xbot(:,2),xc,'spline');
xsurfnew = interp1(xsurf(:,1),xsurf(:,2),xc,'spline');

% Needed to define partitioning
pos_cent_x = xc'; % x-coordinate
pos_cent_z = mean([xbotnew' xsurfnew'],2); % z-coordinate

xbotnew = [xc(2:end-1)' xbotnew(2:end-1)'];
xsurfnew = [xc' xsurfnew'];

% xb defines the full refined boundary profile
xb = [xbotnew; xsurfnew];


%% Determine internal points

% Find which points fall inside the domain
% Remove the points which lie outside
% xi denotes internal points

pol_new = [[0 3.2]; xbotnew; flipud(xsurfnew)];
inPol = inpolygon(xgrid(:,1),xgrid(:,2),pol_new(:,1),pol_new(:,2));
inPol = find(inPol);
xi = xgrid(inPol,:);


%% Determine indistinct points

% Identify points which fall in hfill/2-neighbourhood of each other
% and leave just one copy. Otherwise RBF matrices will be ill-conditioned

% This procedure is necessary because, when adding the boundary points to
% the grid, some grid points may coincide with the boundary. This leads to
% singularities in RBF matrices

indistinct_par = 0.016; %hfill/2;

indremove = [];
indkeep = [];
for i = 1:length(xi)
    for j = 1:length(xb)
        indistinct(j) = norm(xi(i,:)-xb(j,:));
    end
    indistinct_sort = sort(indistinct);
    if indistinct_sort(1) < indistinct_par
        indremove = [indremove i];
    else
        indkeep = [indkeep i];
    end
end

% Remove the indistinct nodes
xi = xi(indkeep,:);


%% Construct the set of computational points

% We combine the internal and the boundary nodes into one set of nodes
% This set does not represent a grid any more, because the boundary nodes
% do not belong to the grid

% The set of points
xe = [xi;xb];

% The number of computational points
Ne = length(xe);

% Define indices which correspond to internal and boundary point in the set   
ind_inter = 1:length(xi); % internal points
ind_bound = length(xi)+1:length(xe); % all boundary points
ind_bot = length(xi)+1:length(xi)+length(xbotnew); % bottom points
ind_surf = length(xi)+length(xbotnew)+1:length(xe); % surface points


%% Compute the derivative of the surface

% The surface derivative is needed for evaluation the horizontal velocity
% field of ice. We use RBF approximation to calculate the derivative

% Find distances between the points on the surface
rc = xcdist(xc',xc','0');

% Compute RBF matrices
Ac = RBFmat(phi,ep,rc,'0',1);
Axc = RBFmat(phi,ep,rc,'1',1)/Ac;

% Compute the derivative
surfder = Axc*xsurfnew(:,2);
surfder = surfder';

% Assign the value of the surface derivative to the corresponding internal
% points
surfderField = repmat(surfder,Nz+1,1);
surfderField = surfderField(:);
surfderField = surfderField(inPol);
surfderField = surfderField(indkeep);

% Define a vector which depends on the surface derivative and comes in the
% right hand side
surfderField_rhs = [surfderField; zeros(length(xb),1)];

surfder_flip = fliplr(surfder);


%% Define the partitioning 

% Define possible positions of center of partitions
% pos_cent_x = xc'; % x-coordinate
% pos_cent_z = mean([xbotnew(:,2) xsurfnew(:,2)],2); % z-coordinate

% Pick out every fifth
center_x = pos_cent_x(4:5:end);
center_z = pos_cent_z(4:5:end);

% Centers of partitions
xm = [center_x center_z];

% Number of partitions
np = length(xm);


%% Define which points fall in which partition

% Check which points go to which partition
for i=1:np
    flagin = hypot(xm(i,1)-xe(:,1),xm(i,2)-xe(:,2)) <= R; 
    ibox(i).indlocpts = find(flagin);
    Ni(i)=length(ibox(i).indlocpts);
end

% Plot the overlapping disks (partitions)
w = xm;%cat(1,ibox.midpt);
th = linspace(0,2*pi,51);
figure;
hold on
plot(xi(:,1),xi(:,2),'ro')
plot(pol_new(:,1),pol_new(:,2),'k-*','linewidth',3)
for i=1:np
    xxx = w(i,1)+1i*w(i,2) + R*exp(1i*th);  
    plot(w(i,1),w(i,2),'ok','markerfacecolor','r','linewidth',2,'markersize',8)
    hold on;   
    plot(xxx,'-k','linewidth',2)
    xlabel('x','fontsize',20); ylabel('y','fontsize',20)
    axis tight
    axis equal 
end
set(gca,'fontsize',20)


%% Compute the partition of unity weights

% Compute the weight functions for each partition using Shepard's method
s=zeros(Ne,1);
sx=zeros(Ne,1);
sy=zeros(Ne,1);
sxx=zeros(Ne,1);
sxy=zeros(Ne,1);
syy=zeros(Ne,1);

for i=1:np
    [psi{i},psix{i},psiy{i},psixx{i},psixy{i},psiyx{i},psiyy{i}] = RBFPUweight2D(xe(ibox(i).indlocpts,:),xm(i,:),R);
    
    s(ibox(i).indlocpts) = s(ibox(i).indlocpts) + psi{i};
    sx(ibox(i).indlocpts)= sx(ibox(i).indlocpts) + psix{i};
    sy(ibox(i).indlocpts)= sy(ibox(i).indlocpts) + psiy{i};
    sxx(ibox(i).indlocpts) = sxx(ibox(i).indlocpts) + psixx{i};
    sxy(ibox(i).indlocpts) = sxy(ibox(i).indlocpts) + psixy{i};
    syy(ibox(i).indlocpts) = syy(ibox(i).indlocpts) + psiyy{i}; 
end  

% Define partition of unity
w0 = zeros(Ne,1);
for i=1:np
    s1 = s(ibox(i).indlocpts);
    s2 = s1.*s1;
    s3 = s2.*s1;
    sx1 = sx(ibox(i).indlocpts);
    sx2 = sx1.*sx1;
    sxx1 = sxx(ibox(i).indlocpts);
    sy1 = sy(ibox(i).indlocpts);
    sy2 = sy1.*sy1;
    syy1 = syy(ibox(i).indlocpts);
    sxy1 = sxy(ibox(i).indlocpts);


    W{i} =  psi{i}./s1;
    Wx{i} = psix{i}./s1 - psi{i}.*sx1./s2;
    Wy{i} = psiy{i}./s1 - psi{i}.*sy1./s2;
    Wxx{i} = -2*psix{i}.*sx1./s2 + psixx{i}./s1 ...
      + psi{i}.*(2*sx2./s3 - sxx1./s2);
    Wxy{i} = psixy{i}./s1 - psix{i}.*sy1./s2 ... 
      - psiy{i}.*sx1./s2 ...
      - psi{i}.*sxy1./s2 ... 
      + 2*psi{i}.*sx1.*sy1./s3;
    Wyy{i} = -2*psiy{i}.*sy1./s2 + psiyy{i}./s1 ... 
       + psi{i}.*(2*sy2./s3 - syy1./s2);
   
   w0(ibox(i).indlocpts) = w0(ibox(i).indlocpts) + W{i};
end


%% RBF matrices

A0 = spalloc(Ne,Ne,sum(Ni.^2));
Ax = spalloc(Ne,Ne,sum(Ni.^2));
Ay = spalloc(Ne,Ne,sum(Ni.^2));
Axx = spalloc(Ne,Ne,sum(Ni.^2));
Ayy = spalloc(Ne,Ne,sum(Ni.^2));
Axy = spalloc(Ne,Ne,sum(Ni.^2));
for i=1:np
    
    xloc = xe(ibox(i).indlocpts,:);
    rc = xcdist(xloc,xloc,1);
    A0_loc  = RBFmat('mq',ep,rc,'0');
    Ax_loc  = RBFmat('mq',ep,rc,'1',1);
    Ay_loc  = RBFmat('mq',ep,rc,'1',2);
    Axx_loc = RBFmat('mq',ep,rc,'2',1);
    Axy_loc = RBFmat('mq',ep,rc,'m2',[1,2]);
    Ayy_loc = RBFmat('mq',ep,rc,'2',2);
    
    Wi = spdiags(W{i},0,Ni(i),Ni(i));
    Wix = spdiags(Wx{i},0,Ni(i),Ni(i));
    Wiy = spdiags(Wy{i},0,Ni(i),Ni(i));
    Wixx = spdiags(Wxx{i},0,Ni(i),Ni(i));
    Wixy = spdiags(Wxy{i},0,Ni(i),Ni(i));
    Wiyy = spdiags(Wyy{i},0,Ni(i),Ni(i));
    
    X = spdiags(xloc(:,1),0,Ni(i),Ni(i));
    Y = spdiags(xloc(:,2),0,Ni(i),Ni(i));
    
    D0 = Wi; 
    Dx = (Wi*Ax_loc + Wix*A0_loc)/A0_loc;
    Dy = (Wi*Ay_loc + Wiy*A0_loc)/A0_loc;

    Dxx = (Wi*Axx_loc + 2*Wix*Ax_loc + Wixx*A0_loc)/A0_loc; 
    Dyy = (Wi*Ayy_loc + 2*Wiy*Ay_loc + Wiyy*A0_loc)/A0_loc;

    Dxy = (Wi*Axy_loc + Wix*Ay_loc + Wiy*Ax_loc + Wixy*A0_loc)/A0_loc;
    
    A0(ibox(i).indlocpts,ibox(i).indlocpts) = A0(ibox(i).indlocpts,ibox(i).indlocpts) + D0;
    Ax(ibox(i).indlocpts,ibox(i).indlocpts) = Ax(ibox(i).indlocpts,ibox(i).indlocpts) + Dx;
    Ay(ibox(i).indlocpts,ibox(i).indlocpts) = Ay(ibox(i).indlocpts,ibox(i).indlocpts) + Dy;
    Axx(ibox(i).indlocpts,ibox(i).indlocpts) = Axx(ibox(i).indlocpts,ibox(i).indlocpts) + Dxx;
    Ayy(ibox(i).indlocpts,ibox(i).indlocpts) = Ayy(ibox(i).indlocpts,ibox(i).indlocpts) + Dyy;
    Axy(ibox(i).indlocpts,ibox(i).indlocpts) = Axy(ibox(i).indlocpts,ibox(i).indlocpts) + Dxy;
    
   
end


%% Nonlenear solver -- Fixed point iteration

% In the nonlinear itration we solve for the horizontal velocity

% Make an initial guess
uold = 0.07*ones(length(xe),1); %Uelm;

% Define the counter
k = 1;

L = spalloc(Ne,Ne,sum(Ni.^2));
Lsurf = spalloc(Ne,Ne,sum(Ni.^2));
grad_eta_x = spalloc(Ne,1,sum(Ni));

err = 1;
while err > tol
    
    % Compute derivatives of the horizontal velocity as 
    % it comes as a multiplier in the viscosity
    dudx = Ax*uold;
    dudy = Ay*uold; 
    
    % Set a critical shear rate
    critshearrate = 10^(-12);
    
    % Compute the viscosity eta
    eta = 0.5*Afact^(-1/3)*(dudx.^2 + 0.25*dudy.^2).^(-1/3);
    
    % Find where the viscosity is smaller than the critical shear rate
    ind_eta = find(abs(dudx.^2 + 0.25*dudy.^2)<critshearrate);
    
    % Replace by the critical shear rate, otherwise singularity
    eta(ind_eta) = 0.5*Afact^(-1/3)*(critshearrate).^(-1/3);   
    
    % Compute the gradient of the viscosity 
    grad_eta_x = -Afact^(-1/3)*((dudx.^2 + 0.25*dudy.^2).^(-4/3))/6 .* ...
        (2*dudx.*(Axx*uold) + 0.5*dudy.*(Axy*uold));
    grad_eta_x(ind_eta) = 0; 
    
    grad_eta_y = -Afact^(-1/3)*((dudx.^2 + 0.25*dudy.^2).^(-4/3))/6 .* ...
        (2*dudx.*(Axy*uold) + 0.5*dudy.*(Ayy*uold));
    grad_eta_y(ind_eta) = 0;
    
    % Make an exception for the first iteration
    if k == 1
        eta = ones(size(eta));
        
        grad_eta_x = zeros(size(eta));
        grad_eta_y = zeros(size(eta));
    end
    
    % Define the discrete spatial differential operator
    L = 4*spdiags(grad_eta_x,0,Ne,Ne)*Ax + 1*spdiags(grad_eta_y,0,Ne,Ne)*Ay + spdiags(eta,0,Ne,Ne)*(4*Axx + 1*Ayy);

    % Define the discrete boundary differential operator at the surface
    Lsurf = -2*spdiags(surfder_flip',0,Nx+1,Nx+1)*Ax(ind_surf,:) + 0.5*Ay(ind_surf,:);
    
    % Define the discrete boundary operator at the base
    % Accorsing to the no slip boundary it is an identity operator
    Lbot = speye(length(ind_bot));

    % Embed these boundary parts into the global system
    L(ind_surf,:) = Lsurf;
    L(ind_bot,:) = 0;
    L(ind_bot,ind_bot) = Lbot;

    % Define the right hand side
    rhs = rho*g*surfderField_rhs;

    % Perform an LU decomposition for a more efficient matrix inversion
    [Lu,Uu] = lu(L);

    % Solve for the horizontal velocity
    vel_x = Uu\(Lu\rhs);
    
    % Compute the error between old and new nonlinear iterations
    err = rms(uold-vel_x);
    
    % Update
    uold = vel_x;

    % Update the counter
    k = k + 1;
    
    % If not converged but exceeded the maximal number of iteration, break
    if k > kmax
        break
    end
    
end

% Number of taken iterations
Niter = k-1;

% Flip the boundary points around so that it is one connected line
% otherwise the triangulation routine is unhappy. 
xtri=xe; utri=vel_x;
xtrib=xe(ind_bound,:); utrib=vel_x(ind_bound,:);
xtrb1=xtrib(1:length(ind_bound)/2-1,:); utrb1=utrib(1:length(ind_bound)/2-1,:);
xtrb2=xtrib(length(ind_bound)/2:end,:); utrb2=utrib(length(ind_bound)/2:end,:);
xtri(ind_bound,:)=[xtrb1; flipud(xtrb2)]; utri(ind_bound,:)=[utrb1; flipud(utrb2)];


% Find triangulation that the solution will be interpolated to
Tri= findTriangulation(xtri(:,1),xtri(:,2),ind_bound);

% Finally, interpolate and the plot of the velocity.
figure;
trisurf(Tri,xtri(:,1),xtri(:,2),utri); view(2);
shading interp;
colormap('cool')
colorbar
title('Horizontal velocity')
set(gca,'fontsize',20)
axis tight



